from .config import load, Mode

javascript_libraries = {
    Mode.DEVELOPMENT: [{
            'url': 'https://cdnjs.cloudflare.com/ajax/libs/react/16.8.4/umd/react.development.js',
            'hash': 'sha256-5qXzTpCme2glpjIMD82N9CvHzot7DsNj+J81wNBYCs8=',
        },{
            'url': 'https://cdnjs.cloudflare.com/ajax/libs/react-dom/16.8.4/umd/react-dom.development.js',
            'hash': 'sha256-zaBIa6kuwEspzOIU5TeB5bATnxmvJFmht9smi5jLH/Q=',
        },{
            'url': 'https://unpkg.com/@material-ui/core@4.4.3/umd/material-ui.development.js',
            'hash': 'sha384-WP+hzP3PWwjXK8aVWENvR3yr/EXFWYxCoJRG78AAXq4+523OrXQPPuLKjSIDpq0c',
        },{
            'url': 'https://unpkg.com/react-beautiful-dnd@10.1.0/dist/react-beautiful-dnd.js',
            'hash': 'sha384-E8lHto1m0Yl8oc4JaZihKNlqmEmnm14731VcvV+KXDKJDdbsmXAWQ3zq+RMnKTHD',
        },{
            'url': 'https://cdn.jsdelivr.net/npm/@fortawesome/fontawesome-svg-core@1.2.15/index.js',
            'hash': 'sha256-rkS272/O+51WOu+puesKtsXK+573Zl5xs8tMNf9vqLg=',
        }, {
            'url': 'https://cdn.jsdelivr.net/npm/@fortawesome/free-solid-svg-icons@5.7.2/index.js',
            'hash': 'sha256-TDDimd5W828ptEg7ng0MeEdiugfFSbMEXXWNPomKgaE=',
        }, {
            'url': 'https://cdn.jsdelivr.net/npm/@fortawesome/free-brands-svg-icons@5.7.2/index.js',
            'hash': 'sha256-zyCrPi/XRBvx1KJzGLb3r7eZLBiwh01F0X1yAKhjuo8=',
        }, {
            'url': 'https://cdn.jsdelivr.net/npm/@fortawesome/free-regular-svg-icons@5.7.2/index.js',
            'hash': 'sha256-Rq5uZFfH7u5qFatjnORtobcHno1kadKPkWu2T+HAINs=',
        }],
    Mode.PRODUCTION: [{
            'url': 'https://cdnjs.cloudflare.com/ajax/libs/react/16.8.4/umd/react.production.min.js',
            'hash': 'sha256-ctUamuIgSCQg1wsh8Iw0QbDXScmyXhjJ6lxYUscC3FA=',
        },{
            'url': 'https://cdnjs.cloudflare.com/ajax/libs/react-dom/16.8.4/umd/react-dom.production.min.js',
            'hash': 'sha256-8uWWG/7CB3OS89Cw67+B++t6w0EMGfQE9C6OGps+Wd8=',
        },{
            'url': 'https://unpkg.com/@material-ui/core@4.4.3/umd/material-ui.production.min.js',
            'hash': 'sha384-Hq0+D81MknxCDnlTkxVY4qebYje4cbxqB5Ru0RuAEoiqIrpt6blMhKRhMdBrudvs',
        },{
            'url': 'https://unpkg.com/react-beautiful-dnd@10.1.0/dist/react-beautiful-dnd.min.js',
             'hash': 'sha384-zo+Gb/fetDb0/fqUY1FIPzktRQi8wZjNWFbhVQREJwAqfX+9g5OkGO8NjTnj6xal',
        },{
            'url': 'https://cdn.jsdelivr.net/npm/@fortawesome/fontawesome-svg-core@1.2.15/index.min.js',
            'hash': 'sha256-EnesNAely1pxccb7VlP95XbVbfOsAcKMgKKi8kjEgw0=',
        }, {
            'url': 'https://cdn.jsdelivr.net/npm/@fortawesome/free-solid-svg-icons@5.7.2/index.min.js',
            'hash': 'sha256-n/Fa8BHquenm3pJEPic5sktYsKlSzXIWY5VT+6/SiCc=',
        }, {
            'url': 'https://cdn.jsdelivr.net/npm/@fortawesome/free-brands-svg-icons@5.7.2/index.min.js',
            'hash': 'sha256-4vKu/atvk2RMilC2lkRpBSQ21wUnsI6Sj4/H7K13iKc=',
        }, {
            'url': 'https://cdn.jsdelivr.net/npm/@fortawesome/free-regular-svg-icons@5.7.2/index.min.js',
            'hash': 'sha256-R2nmSJcsyc4Z36JQJYeVAyT+ZDAWfFfdCb0loPwR29A=',
        }],
}



"""initial

Revision ID: 56aa2b371760
Revises: 
Create Date: 2019-03-05 21:29:06.717677

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '56aa2b371760'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('projects',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(), nullable=False),
    sa.Column('desc', sa.String(), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('users',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('sub', sa.String(), nullable=False),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('issues',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(), nullable=False),
    sa.Column('desc', sa.String(), nullable=True),
    sa.Column('project_id', sa.Integer(), nullable=False),
    sa.Column('user_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['project_id'], ['projects.id'], ),
    sa.ForeignKeyConstraint(['user_id'], ['users.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('sources',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('user_id', sa.Integer(), nullable=False),
    sa.Column('type', sa.Enum('GITLAB', name='sourcetype'), nullable=False),
    sa.Column('config', sa.JSON(), nullable=False),
    sa.ForeignKeyConstraint(['user_id'], ['users.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('time_entries',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('issue_id', sa.Integer(), nullable=True),
    sa.Column('project_id', sa.Integer(), nullable=False),
    sa.Column('user_id', sa.Integer(), nullable=False),
    sa.Column('start', sa.DateTime(), nullable=True),
    sa.Column('end', sa.DateTime(), nullable=True),
    sa.ForeignKeyConstraint(['issue_id'], ['issues.id'], ),
    sa.ForeignKeyConstraint(['project_id'], ['projects.id'], ),
    sa.ForeignKeyConstraint(['user_id'], ['users.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('time_entries')
    op.drop_table('sources')
    op.drop_table('issues')
    op.drop_table('users')
    op.drop_table('projects')
    # ### end Alembic commands ###



"""rename payment field

Revision ID: 35bc88d241d4
Revises: 28e55086b76c
Create Date: 2019-03-17 20:59:02.186415

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '35bc88d241d4'
down_revision = '28e55086b76c'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('payments', sa.Column('created_user_id', sa.Integer(), nullable=False))
    op.drop_constraint('payments_payer_user_id_fkey', 'payments', type_='foreignkey')
    op.create_foreign_key(None, 'payments', 'users', ['created_user_id'], ['id'], ondelete='CASCADE')
    op.drop_column('payments', 'payer_user_id')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('payments', sa.Column('payer_user_id', sa.INTEGER(), autoincrement=False, nullable=False))
    op.drop_constraint(None, 'payments', type_='foreignkey')
    op.create_foreign_key('payments_payer_user_id_fkey', 'payments', 'users', ['payer_user_id'], ['id'], ondelete='CASCADE')
    op.drop_column('payments', 'created_user_id')
    # ### end Alembic commands ###



"""add email to user

Revision ID: 33fd38e158e0
Revises: 56aa2b371760
Create Date: 2019-03-09 16:34:05.377245

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '33fd38e158e0'
down_revision = '56aa2b371760'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('users', sa.Column('email', sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('users', 'email')
    # ### end Alembic commands ###




from enum import Enum
from configpp.tree import Tree, Settings, DatabaseLeaf, PythonLoggerLeaf, NodeBase
from configpp.soil import Config
from voluptuous import Any

from .auth import backends

backend_names = [b.OAUTH_NAME for b in backends]

tree = Tree(Settings(convert_underscores_to_hypens = True, convert_camel_case_to_hypens = True))


class OAuthConfig(NodeBase):

    id = str
    secret = str

    def serialize(self):
        return self.__dict__

class Mode(Enum):

    DEVELOPMENT = 'development'
    PRODUCTION = 'production'

@tree.root()
class PabuConfig():

    database = DatabaseLeaf

    auth = tree.dict_node(Any(*backend_names), OAuthConfig)

    mode = Mode

    logger = dict

def load() -> PabuConfig:
    config_loader = Config('pabu.yaml')

    if not config_loader.load():
        return None

    return tree.load(config_loader.data)



import enum
from datetime import datetime

from sqlalchemy import Column, Integer, String, ForeignKey, Enum, JSON, DateTime, Table
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship


Base = declarative_base()

projects_users = Table('projects_users', Base.metadata,
    Column('project_id', Integer, ForeignKey('projects.id', ondelete = 'CASCADE')),
    Column('user_id', Integer, ForeignKey('users.id', ondelete = 'CASCADE'))
)

issues_tags = Table('issues_tags', Base.metadata,
    Column('issue_id', Integer, ForeignKey('issues.id', ondelete = 'CASCADE')),
    Column('tag_id', Integer, ForeignKey('tags.id', ondelete = 'CASCADE'))
)

class User(Base):
    __tablename__ = 'users'

    id = Column(Integer, primary_key=True)
    sub = Column(String, nullable = False)
    email = Column(String)
    name = Column(String)
    avatar = Column(String)

class Project(Base):
    __tablename__ = 'projects'

    id = Column(Integer, primary_key=True)
    name = Column(String, nullable = False)
    desc = Column(String)

    users = relationship("User", secondary = projects_users)

class SourceType(enum.Enum):

    GITLAB = 'gitlab'

class Source(Base):
    __tablename__ = 'sources'

    id = Column(Integer, primary_key=True)
    user_id = Column(Integer, ForeignKey('users.id', ondelete = 'CASCADE'), nullable = False)
    type = Column(Enum(SourceType), nullable = False)
    config = Column(JSON, nullable = False)

class Issue(Base):
    __tablename__ = 'issues'

    id = Column(Integer, primary_key=True)
    name = Column(String, nullable = False)
    desc = Column(String, default = '')
    project_id = Column(Integer, ForeignKey('projects.id', ondelete = 'CASCADE'), nullable = False)
    reporter_id = Column(Integer, ForeignKey('users.id', ondelete = 'CASCADE'), nullable = False)
    assignee_id = Column(Integer, ForeignKey('users.id', ondelete = 'SET NULL'))
    status = Column(String, default = 'todo', nullable = False)
    rank = Column(Integer, default = 0)
    status_date = Column(DateTime, default = datetime.now, nullable = False)

    project = relationship("Project", backref = "issues")
    tags = relationship("Tag", secondary = issues_tags)

class TimeEntry(Base):
    __tablename__ = 'time_entries'

    id = Column(Integer, primary_key=True)
    issue_id = Column(Integer, ForeignKey('issues.id', ondelete = 'SET NULL'))
    project_id = Column(Integer, ForeignKey('projects.id', ondelete = 'CASCADE'), nullable = False)
    user_id = Column(Integer, ForeignKey('users.id', ondelete = 'CASCADE'), nullable = False)
    start = Column(DateTime, default = datetime.now)
    end = Column(DateTime)

    issue = relationship("Issue", backref = "time_entries")
    project = relationship("Project", backref = "time_entries")

class Payment(Base):
    __tablename__ = 'payments'

    id = Column(Integer, primary_key=True)
    project_id = Column(Integer, ForeignKey('projects.id', ondelete = 'CASCADE'), nullable = False)
    created_user_id = Column(Integer, ForeignKey('users.id', ondelete = 'CASCADE'), nullable = False)
    paid_user_id = Column(Integer, ForeignKey('users.id', ondelete = 'CASCADE'), nullable = False)
    amount = Column(Integer, nullable = False)
    time = Column(DateTime, default = datetime.now)
    note = Column(String, default = '')

    project = relationship("Project", backref = "payments")

class ProjectInvitationToken(Base):

    __tablename__ = 'project_invitation_tokens'

    id = Column(Integer, primary_key = True)
    project_id = Column(Integer, ForeignKey('projects.id', ondelete = 'CASCADE'), nullable = False)
    token = Column(String, nullable = False)

    project = relationship("Project", backref = "tokens")

class Tag(Base):

    __tablename__ = 'tags'

    id = Column(Integer, primary_key = True)
    project_id = Column(Integer, ForeignKey('projects.id', ondelete = 'CASCADE'), nullable = False)
    name = Column(String, nullable = False)

    project = relationship("Project", backref = "tags")



from contextlib import contextmanager

from sqlalchemy.orm import sessionmaker
from sqlalchemy import create_engine


class Database():

    def __init__(self, url: str):
        self._url = url

    @contextmanager
    def session_scope(self, autocommit = True):
        """Provide a transactional scope around a series of operations."""
        session = sessionmaker(bind=create_engine(self._url))()
        try:
            yield session
            if autocommit:
                session.commit()
        except:
            session.rollback()
            raise
        finally:
            session.close()




from flask import session
from loginpass import Google, Twitter, Facebook, GitHub, Dropbox, Reddit, Gitlab, Slack, Discord, Bitbucket, Spotify, Twitch

backends = (Google, Twitter, Facebook, GitHub, Dropbox, Reddit, Gitlab, Slack, Discord, Bitbucket, Spotify, Twitch)

def is_logged_in() -> bool:
    return 'user_info' in session

def get_user_id() -> int:
    return session['user_id']



"""add tags

Revision ID: 3244926da8ec
Revises: e1c315078dc3
Create Date: 2019-09-26 09:34:01.630938

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '3244926da8ec'
down_revision = 'e1c315078dc3'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('tags',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('project_id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(), nullable=False),
    sa.ForeignKeyConstraint(['project_id'], ['projects.id'], ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('issues_tags',
    sa.Column('issue_id', sa.Integer(), nullable=True),
    sa.Column('tag_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['issue_id'], ['issues.id'], ondelete='CASCADE'),
    sa.ForeignKeyConstraint(['tag_id'], ['tags.id'], ondelete='CASCADE')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('issues_tags')
    op.drop_table('tags')
    # ### end Alembic commands ###



"""add ondelete to foreign keys

Revision ID: 9aaefe492ae9
Revises: 65e7825b09b1
Create Date: 2019-03-13 12:31:40.909749

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '9aaefe492ae9'
down_revision = '65e7825b09b1'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint('issues_project_id_fkey', 'issues', type_='foreignkey')
    op.drop_constraint('issues_user_id_fkey', 'issues', type_='foreignkey')
    op.create_foreign_key(None, 'issues', 'projects', ['project_id'], ['id'], ondelete='CASCADE')
    op.create_foreign_key(None, 'issues', 'users', ['user_id'], ['id'], ondelete='CASCADE')
    op.drop_constraint('payments_paid_user_id_fkey', 'payments', type_='foreignkey')
    op.drop_constraint('payments_payer_user_id_fkey', 'payments', type_='foreignkey')
    op.drop_constraint('payments_project_id_fkey', 'payments', type_='foreignkey')
    op.create_foreign_key(None, 'payments', 'users', ['paid_user_id'], ['id'], ondelete='CASCADE')
    op.create_foreign_key(None, 'payments', 'users', ['payer_user_id'], ['id'], ondelete='CASCADE')
    op.create_foreign_key(None, 'payments', 'projects', ['project_id'], ['id'], ondelete='CASCADE')
    op.drop_constraint('project_invitations_project_id_fkey', 'project_invitations', type_='foreignkey')
    op.create_foreign_key(None, 'project_invitations', 'projects', ['project_id'], ['id'], ondelete='CASCADE')
    op.drop_constraint('sources_user_id_fkey', 'sources', type_='foreignkey')
    op.create_foreign_key(None, 'sources', 'users', ['user_id'], ['id'], ondelete='CASCADE')
    op.drop_constraint('time_entries_issue_id_fkey', 'time_entries', type_='foreignkey')
    op.drop_constraint('time_entries_project_id_fkey', 'time_entries', type_='foreignkey')
    op.drop_constraint('time_entries_user_id_fkey', 'time_entries', type_='foreignkey')
    op.create_foreign_key(None, 'time_entries', 'issues', ['issue_id'], ['id'], ondelete='SET NULL')
    op.create_foreign_key(None, 'time_entries', 'projects', ['project_id'], ['id'], ondelete='CASCADE')
    op.create_foreign_key(None, 'time_entries', 'users', ['user_id'], ['id'], ondelete='CASCADE')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, 'time_entries', type_='foreignkey')
    op.drop_constraint(None, 'time_entries', type_='foreignkey')
    op.drop_constraint(None, 'time_entries', type_='foreignkey')
    op.create_foreign_key('time_entries_user_id_fkey', 'time_entries', 'users', ['user_id'], ['id'])
    op.create_foreign_key('time_entries_project_id_fkey', 'time_entries', 'projects', ['project_id'], ['id'])
    op.create_foreign_key('time_entries_issue_id_fkey', 'time_entries', 'issues', ['issue_id'], ['id'])
    op.drop_constraint(None, 'sources', type_='foreignkey')
    op.create_foreign_key('sources_user_id_fkey', 'sources', 'users', ['user_id'], ['id'])
    op.drop_constraint(None, 'project_invitations', type_='foreignkey')
    op.create_foreign_key('project_invitations_project_id_fkey', 'project_invitations', 'projects', ['project_id'], ['id'])
    op.drop_constraint(None, 'payments', type_='foreignkey')
    op.drop_constraint(None, 'payments', type_='foreignkey')
    op.drop_constraint(None, 'payments', type_='foreignkey')
    op.create_foreign_key('payments_project_id_fkey', 'payments', 'projects', ['project_id'], ['id'])
    op.create_foreign_key('payments_payer_user_id_fkey', 'payments', 'users', ['payer_user_id'], ['id'])
    op.create_foreign_key('payments_paid_user_id_fkey', 'payments', 'users', ['paid_user_id'], ['id'])
    op.drop_constraint(None, 'issues', type_='foreignkey')
    op.drop_constraint(None, 'issues', type_='foreignkey')
    op.create_foreign_key('issues_user_id_fkey', 'issues', 'users', ['user_id'], ['id'])
    op.create_foreign_key('issues_project_id_fkey', 'issues', 'projects', ['project_id'], ['id'])
    # ### end Alembic commands ###



"""project_users table

Revision ID: 401b70052c48
Revises: 33fd38e158e0
Create Date: 2019-03-09 18:41:25.371331

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '401b70052c48'
down_revision = '33fd38e158e0'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('projects_users_',
    sa.Column('project_id', sa.Integer(), nullable=True),
    sa.Column('user_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['project_id'], ['projects.id'], ),
    sa.ForeignKeyConstraint(['user_id'], ['users.id'], )
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('projects_users_')
    # ### end Alembic commands ###



"""add status to issue

Revision ID: 6d3439e14660
Revises: 04c84d5e2a21
Create Date: 2019-03-22 07:57:32.625406

"""
from alembic import op
import sqlalchemy as sa
from pabu.tools import get_table

# revision identifiers, used by Alembic.
revision = '6d3439e14660'
down_revision = '04c84d5e2a21'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('issues', sa.Column('status', sa.String(), nullable=True))

    issues = get_table(op.get_bind(), 'issues')
    op.execute(issues.update().values(status = 'todo'))

    op.alter_column('issues', 'status', nullable=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('issues', 'status')
    # ### end Alembic commands ###






"""add ondelete to foreign keys to projects_users

Revision ID: 28e55086b76c
Revises: 9aaefe492ae9
Create Date: 2019-03-13 12:33:09.522274

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '28e55086b76c'
down_revision = '9aaefe492ae9'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint('projects_users_user_id_fkey', 'projects_users', type_='foreignkey')
    op.drop_constraint('projects_users_project_id_fkey', 'projects_users', type_='foreignkey')
    op.create_foreign_key(None, 'projects_users', 'projects', ['project_id'], ['id'], ondelete='CASCADE')
    op.create_foreign_key(None, 'projects_users', 'users', ['user_id'], ['id'], ondelete='CASCADE')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, 'projects_users', type_='foreignkey')
    op.drop_constraint(None, 'projects_users', type_='foreignkey')
    op.create_foreign_key('projects_users_project_id_fkey', 'projects_users', 'projects', ['project_id'], ['id'])
    op.create_foreign_key('projects_users_user_id_fkey', 'projects_users', 'users', ['user_id'], ['id'])
    # ### end Alembic commands ###



import logging
import logging.config
import pkg_resources
import json
import os

from flask import Flask, render_template, session
from werkzeug.wsgi import DispatcherMiddleware
import pkg_resources
from time import time

from .config import load, Mode
from .db import Database
from .models import User
from .controllers.api import add_api_controllers
from .controllers.auth import add_auth_controllers
from .auth import is_logged_in
from .tools import get_all_project_data
from .javascript_libraries import javascript_libraries
from . import CHANGELOG_FILE_PATH

frontend = Flask('pabu')
api = Flask('api')
auth = Flask('auth')

secret_key = 'j32bd4kj2d4jkvgh23jk4hvgj23hg4ekjhg'

frontend.secret_key = secret_key
auth.secret_key = secret_key
api.secret_key = secret_key

wsgi_application = DispatcherMiddleware(frontend, {
    '/api': api,
    '/auth': auth,
})

config = load()

if not config:
    print('no config!')
    exit(1)

for name, cfg in config.auth.items():
    auth.config[(name + '_client_id').upper()] = cfg.id
    auth.config[(name + '_client_secret').upper()] = cfg.secret

logging.config.dictConfig(config.logger)

logger = logging.getLogger(__name__)

logger.info("App start")

db = Database(str(config.database))

changelog = []

if os.path.isfile(CHANGELOG_FILE_PATH):
    with open(CHANGELOG_FILE_PATH) as f:
        changelog = json.load(f)
else:
    if config.mode == Mode.PRODUCTION:
        logger.error("Changelog is missing!")
    else:
        logger.info("Changelog is missing.")

@frontend.route('/')
@frontend.route('/<path:path>')
def index(path = None):

    return render_template('index.html',
        data = {
            'changelog': changelog,
            'userInfo': session.get('user_info'),
            'isLoggedIn': is_logged_in(),
            'authBackendNames': list(config.auth.keys()),
            'version': pkg_resources.get_distribution('pabu').version,
            'initialData': get_all_project_data(db) if is_logged_in() else None,
        },
        dev_mode = config.mode == Mode.DEVELOPMENT,
        version = str(time()) if config.mode == Mode.DEVELOPMENT else pkg_resources.get_distribution("pabu").version,
        javascript_libraries = javascript_libraries[config.mode],
    )

add_api_controllers(api, db)
add_auth_controllers(auth, config, db)



"""rename invite table

Revision ID: 04c84d5e2a21
Revises: 35bc88d241d4
Create Date: 2019-03-18 20:50:05.624146

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '04c84d5e2a21'
down_revision = '35bc88d241d4'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('project_invitation_tokens',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('project_id', sa.Integer(), nullable=False),
    sa.Column('token', sa.String(), nullable=False),
    sa.ForeignKeyConstraint(['project_id'], ['projects.id'], ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('id')
    )
    op.drop_table('project_invitations')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('project_invitations',
    sa.Column('id', sa.INTEGER(), autoincrement=True, nullable=False),
    sa.Column('project_id', sa.INTEGER(), autoincrement=False, nullable=False),
    sa.Column('token', sa.VARCHAR(), autoincrement=False, nullable=False),
    sa.ForeignKeyConstraint(['project_id'], ['projects.id'], name='project_invitations_project_id_fkey', ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('id', name='project_invitations_pkey')
    )
    op.drop_table('project_invitation_tokens')
    # ### end Alembic commands ###



"""add payments table

Revision ID: 65e7825b09b1
Revises: 4d15a831ea33
Create Date: 2019-03-13 09:34:46.148079

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '65e7825b09b1'
down_revision = '4d15a831ea33'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('payments',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('project_id', sa.Integer(), nullable=False),
    sa.Column('payer_user_id', sa.Integer(), nullable=False),
    sa.Column('paid_user_id', sa.Integer(), nullable=False),
    sa.Column('amount', sa.Integer(), nullable=False),
    sa.Column('time', sa.DateTime(), nullable=True),
    sa.Column('note', sa.String(), nullable=True),
    sa.ForeignKeyConstraint(['paid_user_id'], ['users.id'], ),
    sa.ForeignKeyConstraint(['payer_user_id'], ['users.id'], ),
    sa.ForeignKeyConstraint(['project_id'], ['projects.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('project_invitations',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('project_id', sa.Integer(), nullable=False),
    sa.Column('token', sa.String(), nullable=False),
    sa.ForeignKeyConstraint(['project_id'], ['projects.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.add_column('users', sa.Column('name', sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('users', 'name')
    op.drop_table('project_invitations')
    op.drop_table('payments')
    # ### end Alembic commands ###



"""add avatar column to users table

Revision ID: 6fffa1625c78
Revises: 4d5027f9faac
Create Date: 2019-03-24 19:04:18.059991

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '6fffa1625c78'
down_revision = '4d5027f9faac'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('users', sa.Column('avatar', sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('users', 'avatar')
    # ### end Alembic commands ###



"""add rank to issue

Revision ID: 4d5027f9faac
Revises: 6d3439e14660
Create Date: 2019-03-23 13:37:47.267698

"""
from alembic import op
import sqlalchemy as sa
from pabu.tools import get_table

# revision identifiers, used by Alembic.
revision = '4d5027f9faac'
down_revision = '6d3439e14660'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('issues', sa.Column('rank', sa.Integer(), nullable=True))

    issues = get_table(op.get_bind(), 'issues')
    op.execute(issues.update().values(rank = issues.c.id))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('issues', 'rank')
    # ### end Alembic commands ###



from datetime import datetime

from sqlalchemy.ext.automap import automap_base
from sqlalchemy.orm import class_mapper, ColumnProperty
from sqlalchemy import or_
import sqlalchemy.sql.sqltypes
from voluptuous import Schema

from .db import Database
from .models import Project, Issue, Payment, User, ProjectInvitationToken, TimeEntry, projects_users, Tag
from .auth import get_user_id

def get_table(bind, table_name: str):
    """Returns sqlalchemy core table declaration

    Args:
        bind: op.get_bind()
        table_name: str
    """

    base = automap_base()
    base.prepare(bind, reflect = True)
    return getattr(base.classes, table_name).__table__


def sqla_model_to_voluptuous(model):

    mapping = {
        sqlalchemy.sql.sqltypes.Integer: int,
        sqlalchemy.sql.sqltypes.String: str,
        sqlalchemy.sql.sqltypes.DateTime: str,
    }

    schema_dict = {}

    for prop in class_mapper(model).iterate_properties:
        if not isinstance(prop, ColumnProperty):
            continue
        schema_dict[prop.key] = mapping[type(prop.columns[0].type)]

    return Schema(schema_dict, required = False)


def entry_stat_from_list(time_entries):
    stat = {
        'spent': 0,
        'last_entry': 0,
        'count': len(time_entries),
    }
    for entry in time_entries:
        if entry.end and entry.end.timestamp() > stat['last_entry']:
            stat['last_entry'] = entry.end.timestamp()
        if entry.start.timestamp() > stat['last_entry']:
            stat['last_entry'] = entry.start.timestamp()
        stat['spent'] += ((entry.end or datetime.now()) - entry.start).total_seconds()
    return stat

def project_to_dict(project: Project):
    paid = 0
    for payment in project.payments:
        paid += payment.amount
    return {
        'id': project.id,
        'name': project.name,
        'desc': project.desc,
        'time_stat': entry_stat_from_list(project.time_entries),
        'issues': [i.id for i in project.issues],
        'paid': paid,
        'users': [u.id for u in project.users],
        'payments': [p.id for p in project.payments],
        'tokens': [t.id for t in project.tokens],
        'tags': [t.id for t in project.tags],
    }

def issue_to_dict(issue: Issue):
    return {
        'id': issue.id,
        'name': issue.name,
        'desc': issue.desc,
        'status': issue.status,
        'rank': issue.rank,
        'project_id': issue.project_id,
        'reporter_id': issue.reporter_id,
        'assignee_id': issue.assignee_id,
        'time_stat': entry_stat_from_list(issue.time_entries),
        'time_entries': [t.id for t in issue.time_entries],
        'status_date': issue.status_date.timestamp(),
        'tags': [t.id for t in issue.tags],
    }

def tag_to_dict(tag: Tag):
    return {
        'id': tag.id,
        'name': tag.name,
        'project_id': tag.project_id,
    }

def payment_to_dict(payment: Payment):
    return {
        'id': payment.id,
        'project_id': payment.project_id,
        'created_user_id': payment.created_user_id,
        'paid_user_id': payment.paid_user_id,
        'amount': payment.amount,
        'time': payment.time.timestamp(),
        'note': payment.note,
    }

def user_to_dict(user: User):
    return {
        'id': user.id,
        'name': user.name,
        'avatar': user.avatar,
    }

def project_token_to_dict(prj_token: ProjectInvitationToken):
    return {
        'id': prj_token.id,
        'token': prj_token.token,
        'project_id': prj_token.project_id,
    }

def time_entry_to_dict(time_entry: TimeEntry):
    return {
        'id': time_entry.id,
        'issueId': time_entry.issue_id,
        'project_id': time_entry.project_id,
        'userId': time_entry.user_id,
        'start': time_entry.start.timestamp(),
        'end': time_entry.end.timestamp() if time_entry.end else None,
    }


def idize(func, rows: list) -> dict:
    return {r.id: func(r) for r in rows}

def get_all_project_data(db: Database, project_id = None):
    user_id = get_user_id()
    with db.session_scope() as conn:
        project_query = conn.query(Project).join(projects_users).join(User).filter(User.id == user_id)
        if project_id:
            project_query = project_query.filter(Project.id == project_id)
        project_rows = project_query.all()
        project_id_list = [r.id for r in project_rows]

        project_user_rows = conn.query(User).join(projects_users).join(Project).filter(Project.id.in_(project_id_list)).all()
        users = idize(user_to_dict, project_user_rows)

        # if somebody left the project still needs his/her data
        issue_users = [i.assignee_id for i in conn.query(Issue.assignee_id).filter(~Issue.assignee_id.in_(users.keys())).all()]
        issue_users += [i.reporter_id for i in conn.query(Issue.reporter_id).filter(~Issue.reporter_id.in_(users.keys())).all()]

        if issue_users:
            users.update(idize(user_to_dict, conn.query(User).filter(User.id.in_(issue_users))))

        return {
            'projects': idize(project_to_dict, project_rows),
            'issues': idize(issue_to_dict, conn.query(Issue).filter(Issue.project_id.in_(project_id_list)).all()),
            'time_entries': idize(time_entry_to_dict, conn.query(TimeEntry).filter(TimeEntry.project_id.in_(project_id_list)).all()),
            'users': users,
            'payments': idize(payment_to_dict, conn.query(Payment).filter(Payment.project_id.in_(project_id_list)).all()),
            'project_invitation_tokens': idize(project_token_to_dict, conn.query(ProjectInvitationToken).filter(ProjectInvitationToken.project_id.in_(project_id_list)).all()),
            'tags': idize(tag_to_dict, conn.query(Tag).filter(Tag.project_id.in_(project_id_list)).all()),
        }




from flask import Flask, jsonify, session, redirect
from authlib.flask.client import OAuth
from loginpass import create_flask_blueprint
from sqlalchemy import or_, and_

from pabu.auth import backends
from pabu.config import Config
from pabu.db import Database
from pabu.models import User

class Cache(object):
    def __init__(self):
        self._data = {}

    def get(self, k):
        return self._data.get(k)

    def set(self, k, v, timeout=None):
        self._data[k] = v

    def delete(self, k):
        if k in self._data:
            del self._data[k]



def add_auth_controllers(app: Flask, config: Config, db: Database):

    oauth = OAuth(app, Cache())

    def handle_authorize(remote, token, user_info):
        sub = user_info['sub']
        email = user_info['email']
        user_info['providerName'] = remote.name
        session['user_info'] = user_info
        with db.session_scope() as conn:
            user = conn.query(User).filter(or_(User.sub == sub, and_(User.email == email, User.email != None))).first()
            if not user:
                user = User(sub = sub, email = email)
                conn.add(user)
            user.email = email
            user.name = user_info['name']
            user.avatar = user_info['picture']
            conn.flush()
            session['user_id'] = user.id

        return redirect('/')

    for backend in backends:
        bp = create_flask_blueprint(backend, oauth, handle_authorize)
        app.register_blueprint(bp, url_prefix='/{}'.format(backend.OAUTH_NAME))


    @app.route('/logout')
    def logout(): # pylint: disable=unused-variable
        del session['user_info']
        del session['user_id']
        return redirect('/')



import logging
import random
import string
from datetime import datetime, timedelta
from functools import partial

from dateutil.parser import parse as dateparser
from flask import Flask, abort, request
from jsonrpc.backend.flask import JSONRPCAPI
from pytimeparse import parse
from voluptuous import Schema

from pabu.auth import get_user_id, is_logged_in
from pabu.db import Database
from pabu.models import Issue, Payment, Project, ProjectInvitationToken, TimeEntry, User, projects_users, Tag
from pabu.tools import (entry_stat_from_list, issue_to_dict, project_to_dict, project_token_to_dict, sqla_model_to_voluptuous, tag_to_dict,
                        time_entry_to_dict, user_to_dict, payment_to_dict, get_all_project_data as get_all_project_data_)

logger = logging.getLogger(__name__)

def random_str(len: int):
    return ''.join(random.choice(string.ascii_lowercase + string.digits) for _ in range(len))

def add_api_controllers(app: Flask, db: Database):

    jsonrpc_api = JSONRPCAPI()

    app.add_url_rule('/', 'api', jsonrpc_api.as_view(), methods=['POST'])

    def get_current_user(conn):
        user_id = get_user_id()
        return conn.query(User).filter(User.id == user_id).first()

    @app.before_request
    def check_login(): # pylint: disable=unused-variable
        if not is_logged_in():
            abort(401)

    @app.before_request
    def common_log(): # pylint: disable=unused-variable
        logger.info("JSONRPC method '%s' called with args: %s" % (request.json['method'], request.json['params']))
        request.start = datetime.now()

    @app.after_request
    def print_time(response): # pylint: disable=unused-variable
        logger.info("JSONRPC method '%s' responded in %s ms", request.json['method'], (datetime.now() - request.start).total_seconds()*1000)
        return response

    def check_project(project_id, conn):
        user_id = get_user_id()
        if conn.query(Project).join(projects_users).join(User).filter(User.id == user_id).filter(Project.id == project_id).count() != 1:
            logger.error("check_project failed! project: %s, user: %s", project_id, user_id)
            abort(404)

    @jsonrpc_api.dispatcher.add_method
    def create_project(name: str, desc: str): # pylint: disable=unused-variable
        with db.session_scope() as conn:
            prj = Project(name = name, desc = desc, users = [get_current_user(conn)])
            conn.add(prj)
            conn.flush()
            return project_to_dict(prj)

    @jsonrpc_api.dispatcher.add_method
    def update_project(id: int, name: str, desc: str): # pylint: disable=unused-variable
        with db.session_scope() as conn:
            check_project(id, conn)
            project = conn.query(Project).filter(Project.id == id).first() # type: Project
            project.name = name
            project.desc = desc
            return project_to_dict(project)

    @jsonrpc_api.dispatcher.add_method
    def get_projects(id: int = None): # pylint: disable=unused-variable
        user_id = get_user_id()
        with db.session_scope() as conn:
            qs = conn.query(Project).outerjoin(Payment).join(projects_users).join(User).filter(User.id == user_id) \
                     .outerjoin(Issue,Project.id == Issue.project_id)
            if id:
                qs = qs.filter(Project.id == id)
            data = {r.id: project_to_dict(r) for r in qs.all()}
        return data

    def process_resources(raw_data_list: list, model, conn, handler):
        schema = sqla_model_to_voluptuous(model)
        res = []

        for raw_data in raw_data_list:

            if 'id' in raw_data and raw_data['id'] is None:
                del raw_data['id']

            data = schema(raw_data)

            if 'pre_process' in handler:
                data.update(handler['pre_process'](data))

            if 'id' in data:
                instance = conn.query(model).filter(model.id == data['id']).first()
                if not instance:
                    abort(404)
                handler['checker'](instance)
                if 'pre_update' in handler:
                    handler['pre_update'](data, instance)
                for key, value in data.items():
                    setattr(instance, key, value)
            else:
                instance = model(**data)
                handler['checker'](instance)
                conn.add(instance)

            conn.flush()

            if 'post_process' in handler:
                new_data = handler['post_process'](data, instance)
                if new_data:
                    for key, value in new_data.items():
                        setattr(instance, key, value)
                    conn.flush()

            res.append(instance)

        return res

    @jsonrpc_api.dispatcher.add_method
    def process_issues(issues): # pylint: disable=unused-variable
        def pre_update(new_data: dict, old_issue: Issue):
            if 'status' in new_data and old_issue.status != new_data['status']:
                new_data['status_date'] = datetime.now()
                if new_data['status'] == 'in progress':
                    new_data['assignee_id'] = get_user_id()
                if new_data['status'] == 'todo':
                    new_data['assignee_id'] = None

        with db.session_scope() as conn:
            issues = process_resources(issues, Issue, conn, {
                'pre_process': lambda d: {'reporter_id': get_user_id()},
                'pre_update': pre_update,
                'checker': lambda i: check_project(i.project_id, conn),
                'post_process': lambda d, i: None if 'id' in d else {'rank': i.id},
            })
            return {i.id: issue_to_dict(i) for i in issues}

    @jsonrpc_api.dispatcher.add_method
    def process_tags(issue_id, tags): # pylint: disable=unused-variable
        with db.session_scope() as conn:
            project_id = conn.query(Project.id).join(Issue, Issue.project_id == Project.id).filter(Issue.id == issue_id).first().id
            check_project(project_id, conn)
            existing_tags = conn.query(Tag).filter(Tag.project_id == project_id).filter(Tag.name.in_(tags)).all()
            existing_tag_names = set(t.name for t in existing_tags)
            new_tags = [Tag(name = n, project_id = project_id) for n in list(set(tags) - existing_tag_names)]
            if new_tags:
                conn.add(*new_tags)
            issue = conn.query(Issue).filter(Issue.id == issue_id).first()
            issue.tags = new_tags + existing_tags

    @jsonrpc_api.dispatcher.add_method
    def get_tags(project_id: int): # pylint: disable=unused-variable
        user_id = get_user_id()
        with db.session_scope() as conn:
            rows = conn.query(Tag).join(Project).join(projects_users).join(User).filter(User.id == user_id) \
                       .filter(Project.id == project_id).all()
            return {r.id: tag_to_dict(r) for r in rows}


    @jsonrpc_api.dispatcher.add_method
    def get_issues(project_id: int): # pylint: disable=unused-variable
        user_id = get_user_id()
        with db.session_scope() as conn:
            rows = conn.query(Issue).join(Project).join(projects_users).join(User).filter(User.id == user_id) \
                       .filter(Project.id == project_id).all()
            return {r.id: issue_to_dict(r) for r in rows}

    @jsonrpc_api.dispatcher.add_method
    def get_time_entries(project_id: int): # pylint: disable=unused-variable
        with db.session_scope() as conn:
            check_project(project_id, conn)
            rows = conn.query(TimeEntry).filter(Project.id == project_id).all()
            return {r.id: time_entry_to_dict(r) for r in rows}

    @jsonrpc_api.dispatcher.add_method
    def add_time(project_id: int, amount: str, start: str, issue_id = None): # pylint: disable=unused-variable
        user_id = get_user_id()
        with db.session_scope() as conn:
            check_project(project_id, conn)
            seconds = parse(amount)
            start = dateparser(start) if start else (datetime.now() - timedelta(seconds = seconds))
            end = start + timedelta(seconds = seconds)
            entry = TimeEntry(project_id = project_id, issue_id = issue_id, start = start, end = end, user_id = user_id)
            conn.add(entry)

    @jsonrpc_api.dispatcher.add_method
    def start_time(project_id, issue_id = None): # pylint: disable=unused-variable
        user_id = get_user_id()
        with db.session_scope() as conn:
            check_project(project_id, conn)
            entry = TimeEntry(project_id = project_id, issue_id = issue_id, user_id = user_id)
            conn.add(entry)

    @jsonrpc_api.dispatcher.add_method
    def stop_time(): # pylint: disable=unused-variable
        user_id = get_user_id()
        with db.session_scope() as conn:
            entry = conn.query(TimeEntry).filter(TimeEntry.user_id == user_id).filter(TimeEntry.end.is_(None)).first()
            if entry:
                entry.end = datetime.now()
                return time_entry_to_dict(entry)

    @jsonrpc_api.dispatcher.add_method
    def get_ticking_stat(): # pylint: disable=unused-variable
        user_id = get_user_id()
        with db.session_scope() as conn:
            entry = conn.query(TimeEntry).filter(TimeEntry.user_id == user_id, TimeEntry.end.is_(None)).first()
            if entry:
                return {'ticking': True, 'entry': time_entry_to_dict(entry)}
            else:
                return {'ticking': False}
        return {}

    @jsonrpc_api.dispatcher.add_method
    def get_payments(project_id: int): # pylint: disable=unused-variable
        with db.session_scope() as conn:
            check_project(project_id, conn)
            rows = conn.query(Payment).filter(Payment.project_id == project_id).all()
            return {r.id:payment_to_dict(r) for r in rows}

    @jsonrpc_api.dispatcher.add_method
    def add_payment(project_id: int, amount: str, paid_user_id: int, time: str, note: str): # pylint: disable=unused-variable
        user_id = get_user_id()
        with db.session_scope() as conn:
            check_project(project_id, conn)
            payment = Payment(project_id = project_id, amount = parse(amount), paid_user_id = paid_user_id, created_user_id = user_id,
                              note = note, time = dateparser(time) if time else datetime.now())
            conn.add(payment)
            return True

    @jsonrpc_api.dispatcher.add_method
    def get_project_users(project_id: int): # pylint: disable=unused-variable
        with db.session_scope() as conn:
            check_project(project_id, conn)
            rows = conn.query(User).join(projects_users).join(Project).filter(Project.id == project_id).all()
            return {r.id: user_to_dict(r) for r in rows}

    @jsonrpc_api.dispatcher.add_method
    def delete_project(id: int): # pylint: disable=unused-variable
        with db.session_scope() as conn:
            check_project(id, conn)
            conn.query(Project).filter(Project.id == id).delete()
            return True

    @jsonrpc_api.dispatcher.add_method
    def delete_issue(id: int): # pylint: disable=unused-variable
        user_id = get_user_id()
        with db.session_scope() as conn:
            issue = conn.query(Issue).join(Project).join(projects_users).join(User).filter(User.id == user_id).filter(Issue.id == id).first()
            if not issue:
                abort(404)
            conn.delete(issue)
            return True

    @jsonrpc_api.dispatcher.add_method
    def delete_time_entry(id: int): # pylint: disable=unused-variable
        user_id = get_user_id()
        with db.session_scope() as conn:
            qs = conn.query(TimeEntry).join(Project).join(projects_users).join(User).filter(User.id == user_id).filter(TimeEntry.id == id)
            entry = qs.first()
            if not entry:
                abort(404)
            conn.delete(entry)
            return True

    @jsonrpc_api.dispatcher.add_method
    def delete_payment(id: int): # pylint: disable=unused-variable
        user_id = get_user_id()
        with db.session_scope() as conn:
            qs = conn.query(Payment).join(Project).join(projects_users).join(User).filter(User.id == user_id).filter(Payment.id == id)
            payment = qs.first()
            if not payment:
                abort(404)
            conn.delete(payment)
            return True

    @jsonrpc_api.dispatcher.add_method
    def create_project_token(id: int): # pylint: disable=unused-variable
        with db.session_scope() as conn:
            proj_token = ProjectInvitationToken(project_id = id, token = random_str(32))
            conn.add(proj_token)
            conn.flush()
            return project_token_to_dict(proj_token)

    @jsonrpc_api.dispatcher.add_method
    def delete_project_token(id: int): # pylint: disable=unused-variable
        user_id = get_user_id()
        with db.session_scope() as conn:
            qs = conn.query(ProjectInvitationToken).join(Project).join(projects_users).join(User).filter(User.id == user_id) \
                      .filter(ProjectInvitationToken.id == id)
            prj_token = qs.first()
            if not prj_token:
                abort(404)
            conn.delete(prj_token)
            return True

    @jsonrpc_api.dispatcher.add_method
    def get_project_tokens(id: int): # pylint: disable=unused-variable
        with db.session_scope() as conn:
            check_project(id, conn)
            rows = conn.query(ProjectInvitationToken).filter(ProjectInvitationToken.project_id == id).all()
            return {r.id: project_token_to_dict(r) for r in rows}

    @jsonrpc_api.dispatcher.add_method
    def join_to_project(token: str): # pylint: disable=unused-variable
        with db.session_scope() as conn:
            user_id = get_user_id()
            prj_token = conn.query(ProjectInvitationToken).filter(ProjectInvitationToken.token == token).first()
            if not prj_token:
                abort(404)
            project = conn.query(Project).filter(Project.id == prj_token.project_id).first() # type: Project
            user = conn.query(User).filter(User.id == user_id).first()
            if user not in project.users:
                project.users.append(user)
            return True

    @jsonrpc_api.dispatcher.add_method
    def kick_user_from_project(project_id: int, user_id: int): # pylint: disable=unused-variable
        with db.session_scope() as conn:
            check_project(project_id, conn)
            project = conn.query(Project).filter(Project.id == project_id).first() # type: Project
            user = conn.query(User).filter(User.id == user_id).first()
            project.users.remove(user)
            return True

    @jsonrpc_api.dispatcher.add_method
    def leave_project(project_id: int): # pylint: disable=unused-variable
        with db.session_scope() as conn:
            check_project(project_id, conn)
            user_id = get_user_id()
            project = conn.query(Project).filter(Project.id == project_id).first() # type: Project
            user = conn.query(User).filter(User.id == user_id).first()
            project.users.remove(user)
            return True


    @jsonrpc_api.dispatcher.add_method
    def get_all_project_data(id: int): # pylint: disable=unused-variable
        return get_all_project_data_(db, id)

    @jsonrpc_api.dispatcher.add_method
    def ping(): # pylint: disable=unused-variable
        return 1/0



"""add assignee

Revision ID: e1c315078dc3
Revises: 04115bd903c0
Create Date: 2019-03-30 17:48:04.431648

"""
from alembic import op
import sqlalchemy as sa
from pabu.tools import get_table

# revision identifiers, used by Alembic.
revision = 'e1c315078dc3'
down_revision = '04115bd903c0'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('issues', sa.Column('assignee_id', sa.Integer(), nullable=True))
    op.add_column('issues', sa.Column('reporter_id', sa.Integer(), nullable=True))
    op.drop_constraint('issues_user_id_fkey', 'issues', type_='foreignkey')
    op.create_foreign_key(None, 'issues', 'users', ['assignee_id'], ['id'], ondelete='SET NULL')
    op.create_foreign_key(None, 'issues', 'users', ['reporter_id'], ['id'], ondelete='CASCADE')

    issues = get_table(op.get_bind(), 'issues')
    op.execute(issues.update().values(reporter_id = issues.c.user_id))
    op.execute(issues.update().values(assignee_id = issues.c.user_id).where(issues.c.status.in_(['done', 'in progress'])))

    op.alter_column('issues', 'reporter_id', nullable=False)
    op.drop_column('issues', 'user_id')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('issues', sa.Column('user_id', sa.INTEGER(), autoincrement=False, nullable=True))
    op.drop_constraint(None, 'issues', type_='foreignkey')
    op.drop_constraint(None, 'issues', type_='foreignkey')
    op.create_foreign_key('issues_user_id_fkey', 'issues', 'users', ['user_id'], ['id'], ondelete='CASCADE')

    issues = get_table(op.get_bind(), 'issues')
    op.execute(issues.update().values(user_id = issues.c.reporter_id))

    op.drop_column('issues', 'reporter_id')
    op.drop_column('issues', 'assignee_id')
    op.alter_column('issues', 'user_id', nullable=False)
    # ### end Alembic commands ###



"""add status_date to issue

Revision ID: 04115bd903c0
Revises: 6fffa1625c78
Create Date: 2019-03-25 19:35:58.977873

"""
from alembic import op
import sqlalchemy as sa
from pabu.tools import get_table
from datetime import datetime

# revision identifiers, used by Alembic.
revision = '04115bd903c0'
down_revision = '6fffa1625c78'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('issues', sa.Column('status_date', sa.DateTime(), nullable=True))

    issues = get_table(op.get_bind(), 'issues')
    op.execute(issues.update().values(status_date = datetime.now()))

    op.alter_column('issues', 'status_date', nullable=False)

def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('issues', 'status_date')
    # ### end Alembic commands ###



import json
import re
import os
from git import Repo
from collections import OrderedDict

from . import CHANGELOG_FILE_PATH

def generate_changelog():

    r = Repo('.')

    pattern = re.compile('(enh|fix): (.+)')

    data = OrderedDict()

    last_tag = None

    commit_hash_to_tag = {str(t.commit): (t.tag.tag, t.tag.tagged_date) for t in r.tags}

    for commit in r.iter_commits():

        commit_hex = commit.binsha.hex()

        if commit_hex in commit_hash_to_tag:
            (last_tag, last_tag_date) = commit_hash_to_tag[commit_hex]
            data[last_tag] = {'enh': [], 'fix': [], 'date': last_tag_date}

        if not last_tag:
            continue

        matches = pattern.match(commit.message.strip())
        if not matches:
            continue

        prefix, message = matches.groups()

        data[last_tag][prefix].append(message)

    with open(CHANGELOG_FILE_PATH, 'w') as f:
        json.dump([{'name': tag_name, **tag_data} for tag_name, tag_data in data.items()], f, indent=2)
